# -*- time-stamp-pattern: "changed[\s]+:[\s]+%%$"; -*-
# AUTHOR INFORMATION ##########################################################
# file    : cyclical_feature_encoder.py
# author  : Marcel Arpogaus <marcel dot arpogaus at gmail dot com>
#
# created : 2022-01-07 09:02:38 (Marcel Arpogaus)
# changed : 2024-02-19 12:53:30 (Marcel Arpogaus)
# DESCRIPTION #################################################################
# ...
# LICENSE #####################################################################
# Copyright 2022 Marcel Arpogaus
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###############################################################################
from typing import Any, Callable, Tuple

import numpy as np
import pandas as pd


def default_cycl_getter(df: pd.DataFrame, k: str) -> Any:
    """Get the cyclical feature 'k' from a DataFrame or its index.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame object.
    k : str
        Key to get the cyclical feature.

    Returns
    -------
    Any
        Cyclical feature corresponding to 'k'.

    """
    try:
        cycl = getattr(df, k)
    except AttributeError:
        cycl = getattr(df.index, k)
    return cycl


class CyclicalFeatureEncoder:
    def __init__(
        self,
        cycl_name: str,
        cycl_max: int,
        cycl_min: int = 0,
        cycl_getter: Callable = default_cycl_getter,
    ) -> None:
        """Initialize the CyclicalFeatureEncoder object.

        Parameters
        ----------
        cycl_name : str
            Name of the cyclical feature.
        cycl_max : int
            Maximum value of the cyclical feature.
        cycl_min : int, optional
            Minimum value of the cyclical feature, by default 0.
        cycl_getter : Callable, optional
            Function to get the cyclical feature, by default default_cycl_getter.

        """
        self.cycl_max = cycl_max
        self.cycl_min = cycl_min
        self.cycl_getter = cycl_getter
        self.cycl_name = cycl_name

    def encode(self, data: pd.DataFrame) -> Tuple[np.ndarray, np.ndarray]:
        """Encode the cyclical feature into sine and cosine components.

        Parameters
        ----------
        data : pd.DataFrame
            Data containing the cyclical feature.

        Returns
        -------
        Tuple[np.ndarray, np.ndarray]
            Sine and cosine components of the encoded cyclical feature.

        """
        cycl = self.cycl_getter(data, self.cycl_name)
        sin = np.sin(
            2 * np.pi * (cycl - self.cycl_min) / (self.cycl_max - self.cycl_min + 1)
        )
        cos = np.cos(
            2 * np.pi * (cycl - self.cycl_min) / (self.cycl_max - self.cycl_min + 1)
        )
        assert np.allclose(cycl, self.decode(sin, cos)), (
            'Decoding failed. Is "cycl_min/max"'
            f"({self.cycl_min}/{self.cycl_max}) correct?"
        )
        return sin, cos

    def __call__(self, data: pd.DataFrame) -> pd.DataFrame:
        """Call method to encode the cyclical feature and add the sine and cosine
        components as new columns in the data.

        Parameters
        ----------
        data : pd.DataFrame
            Data containing the cyclical feature.

        Returns
        -------
        pd.DataFrame
            Updated data with sine and cosine components.

        """
        data = data.copy()
        sin, cos = self.encode(data)
        data[self.cycl_name + "_sin"] = sin
        data[self.cycl_name + "_cos"] = cos
        return data

    def decode(self, sin: np.ndarray, cos: np.ndarray) -> np.ndarray:
        """Decode the encoded sine and cosine components back to the cyclical feature.

        Parameters
        ----------
        sin : np.ndarray
            Sine component of the encoded cyclical feature.
        cos : np.ndarray
            Cosine component of the encoded cyclical feature.

        Returns
        -------
        np.ndarray
            Decoded cyclical feature.

        """
        angle = (np.arctan2(sin, cos) + 2 * np.pi) % (2 * np.pi)
        return (angle * (self.cycl_max - self.cycl_min + 1)) / (
            2 * np.pi
        ) + self.cycl_min



# -*- time-stamp-pattern: "changed[\s]+:[\s]+%%$"; -*-
# AUTHOR INFORMATION ##########################################################
# file    : groupby_dataset_generator.py
# author  : Marcel Arpogaus <marcel dot arpogaus at gmail dot com>
#
# created : 2022-01-07 09:02:38 (Marcel Arpogaus)
# changed : 2024-02-19 12:57:02 (Marcel Arpogaus)
# DESCRIPTION #################################################################
# ...
# LICENSE #####################################################################
# Copyright 2022 Marcel Arpogaus
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###############################################################################

from typing import Callable, List

import pandas as pd
import tensorflow as tf


class GroupbyDatasetGenerator:
    """Creates a generator that yields the data for each group in the given column.

    Parameters
    ----------
    groupby : str
        The column name to group the dataframe by.
    columns : List[str]
        The list of columns to keep in the dataset.
    dtype : tf.dtypes.DType, optional
        The dtype of the output tensors, by default tf.float32.
    shuffle : bool, optional
        Whether to shuffle the dataset or not, by default False.
    test_mode : bool, optional
        Whether the generator is in test mode or not, by default False.

    """

    def __init__(
        self,
        groupby: str,
        columns: List[str],
        dtype: tf.dtypes.DType = tf.float32,
        shuffle: bool = False,
        test_mode: bool = False,
    ) -> None:
        self.groupby: str = groupby
        self.columns: List[str] = sorted(list(set(columns)))
        self.dtype: tf.dtypes.DType = dtype
        self.shuffle: bool = shuffle
        self.test_mode: bool = test_mode

    def get_generator(self, df: pd.DataFrame) -> Callable[[], tf.Tensor]:
        """Returns the dataset generator with the given parameters.

        Parameters
        ----------
        df : pd.DataFrame
            The source DataFrame to generate data from.

        Returns
        -------
        Callable[[], tf.Tensor]
            A function that when called returns a generator yielding batches of
            data as tensors.

        """
        df.sort_index(inplace=True)
        if self.test_mode:
            ids = df[self.groupby].unique()
            ids = ids[:2]
            df = df[df[self.groupby].isin(ids)]

        grpd = df.groupby(self.groupby)

        def generator() -> tf.Tensor:
            for _, d in grpd:
                yield d[self.columns].values.astype(self.dtype.as_numpy_dtype)

        return generator

    def __call__(self, df: pd.DataFrame) -> tf.data.Dataset:
        """Makes the class instance callable and returns a TensorFlow dataset.

        Parameters
        ----------
        df : pd.DataFrame
            The DataFrame to create the dataset from.

        Returns
        -------
        tf.data.Dataset
            A TensorFlow dataset created from the DataFrame.

        """
        ds: tf.data.Dataset = tf.data.Dataset.from_generator(
            self.get_generator(df),
            output_signature=(
                tf.TensorSpec(shape=[None, len(self.columns)], dtype=self.dtype)
            ),
        )
        if self.shuffle:
            len_ids = df[self.groupby].unique().size
            ds = ds.shuffle(len_ids)
        return ds



# -*- time-stamp-pattern: "changed[\s]+:[\s]+%%$"; -*-
# AUTHOR INFORMATION ##########################################################
# file    : windowed_time_series_pipeline.py
# author  : Marcel Arpogaus <marcel dot arpogaus at gmail dot com>
#
# created : 2022-01-07 09:02:38 (Marcel Arpogaus)
# changed : 2024-02-19 12:53:06 (Marcel Arpogaus)
# DESCRIPTION #################################################################
# ...
# LICENSE #####################################################################
# Copyright 2022 Marcel Arpogaus
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###############################################################################
# REQUIRED PYTHON MODULES #####################################################
from typing import List, Union

import tensorflow as tf
from tensorflow.data import Dataset

from tensorflow_time_series_dataset.pipeline.patch_generator import PatchGenerator
from tensorflow_time_series_dataset.pipeline.patch_processor import PatchPreprocessor


class WindowedTimeSeriesPipeline:
    """A Pipeline to process time-series data in a rolling window for TensorFlow models.

    This class provides functionality to generate windowed datasets from time series
    data for training, validating, and testing machine learning models. It allows for
    the generation of datasets that include historical data windows along with
    corresponding future values (for either regression or classification tasks).

    Parameters
    ----------
    history_size : int
        The size of the history window.
    prediction_size : int
        The size of the prediction window.
    history_columns : List[str]
        The names of the columns to be used for the history window.
    meta_columns : List[str]
        The names of the columns to be used for metadata in each window.
    prediction_columns : List[str]
        The names of the columns from which the future values are predicted.
    shift : int
        The shift (in time units) for the sliding window.
    batch_size : int
        The size of the batch for each dataset iteration.
    cycle_length : int
        The number of data files that the dataset can process in parallel.
    shuffle_buffer_size : int
        The buffer size for data shuffling.
    cache : Union[str, bool]
        Whether to cache the dataset in memory or to a specific file.
    drop_remainder : bool
        Whether to drop the remainder of batches that are not equal to the batch size.

    Raises
    ------
    AssertionError
        If the prediction_size is not greater than zero.

    """

    def __init__(
        self,
        history_size: int,
        prediction_size: int,
        history_columns: List[str],
        meta_columns: List[str],
        prediction_columns: List[str],
        shift: int,
        batch_size: int,
        cycle_length: int,
        shuffle_buffer_size: int,
        cache: Union[str, bool],
        drop_remainder: bool,
    ) -> None:
        assert (
            prediction_size > 0
        ), "prediction_size must be a positive integer greater than zero"
        self.history_size = history_size
        self.prediction_size = prediction_size
        self.window_size = history_size + prediction_size
        self.history_columns = history_columns
        self.meta_columns = meta_columns
        self.prediction_columns = prediction_columns
        self.shift = shift
        self.batch_size = batch_size
        self.cycle_length = cycle_length
        self.shuffle_buffer_size = shuffle_buffer_size
        self.cache = cache
        self.drop_remainder = drop_remainder

    def __call__(self, ds: Dataset) -> Dataset:
        """Applies the pipeline operations to the given dataset.

        Parameters
        ----------
        ds : Dataset
            The input dataset to transform.

        Returns
        -------
        Dataset
            The transformed dataset, ready for model training or evaluation.

        """
        ds = ds.interleave(
            PatchGenerator(self.window_size, self.shift),
            cycle_length=self.cycle_length,
            num_parallel_calls=tf.data.experimental.AUTOTUNE,
        )

        ds = ds.map(
            PatchPreprocessor(
                self.history_size,
                self.history_columns,
                self.meta_columns,
                self.prediction_columns,
            ),
            num_parallel_calls=tf.data.experimental.AUTOTUNE,
        )

        if self.cache:
            ds = ds.cache()

        if self.shuffle_buffer_size:
            ds = ds.shuffle(self.shuffle_buffer_size)

        ds = ds.batch(self.batch_size, drop_remainder=self.drop_remainder)

        ds = ds.prefetch(tf.data.experimental.AUTOTUNE)

        return ds



# -*- time-stamp-pattern: "changed[\s]+:[\s]+%%$"; -*-
# AUTHOR INFORMATION ##########################################################
# file    : time_series_k_fold.py
# author  : Marcel Arpogaus <marcel dot arpogaus at gmail dot com>
#
# created : 2022-01-07 09:02:38 (Marcel Arpogaus)
# changed : 2024-02-19 12:35:49 (Marcel Arpogaus)
# DESCRIPTION #################################################################
# ...
# LICENSE #####################################################################
# Copyright 2022 Marcel Arpogaus
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###############################################################################
# REQUIRED PYTHON MODULES #####################################################
from typing import Dict, List

import numpy as np
import pandas as pd


class TimeSeriesKFold:
    """Time series cross-validation using a sliding window.

    Parameters
    ----------
    fold : int
        The current fold to be used.
    n_folds : int, optional
        The total number of folds (the default is 20).

    """

    def __init__(self, fold: int, n_folds: int = 20) -> None:
        self.n_folds: int = n_folds
        self.fold: int = fold

    def __call__(self, data: pd.DataFrame) -> pd.DataFrame:
        """Divide the data into train/test sets for the current fold.

        Parameters
        ----------
        data : pd.DataFrame
            The data to be folded, requiring a datetime index.

        Returns
        -------
        pd.DataFrame
            The subset of data corresponding to the current fold.

        """
        days: pd.DatetimeIndex = pd.date_range(
            data.index.min(), data.index.max(), freq="D"
        )
        fold_idx: List[np.ndarray] = np.array_split(days.to_numpy(), self.n_folds)
        folds: Dict[int, List[str]] = {
            f: (idx[[0, -1]].astype("datetime64[m]") + np.array([0, 60 * 24 - 1]))
            .astype(str)
            .tolist()
            for f, idx in enumerate(fold_idx)
        }

        return data[folds[self.fold][0] : folds[self.fold][1]]



# -*- time-stamp-pattern: "changed[\s]+:[\s]+%%$"; -*-
# AUTHOR INFORMATION ##########################################################
# file    : visualisation.py
# author  : Marcel Arpogaus <marcel dot arpogaus at gmail dot com>
#
# created : 2022-01-07 16:50:58 (Marcel Arpogaus)
# changed : 2024-02-19 12:07:12 (Marcel Arpogaus)
# DESCRIPTION #################################################################
# ...
# LICENSE #####################################################################
# ...
###############################################################################
from typing import Any, List, Optional, Tuple

import matplotlib.pyplot as plt
import numpy as np
from tensorflow.data import Dataset


def plot_patch(
    ds: Dataset,
    history_size: int,
    prediction_size: int,
    history_columns: List[str],
    prediction_columns: List[str],
    meta_columns: Optional[List[str]] = None,
    figsize: Tuple[int, int] = (16, 8),
    **kwargs: Any,
) -> plt.Figure:
    """Plot history and prediction data from a dataset.

    Parameters
    ----------
    ds : Dataset
        Dataset to extract data from.
    history_size : int
        Number of timesteps in the history.
    prediction_size : int
        Number of timesteps in the prediction.
    history_columns : List[str]
        Columns to use for history data.
    prediction_columns : List[str]
        Columns to use for prediction data.
    meta_columns : List[str], optional
        Columns for additional meta data. Defaults to [].
    figsize : Tuple[int, int], optional
        Size of the figure. Defaults to (16, 8).

    Returns
    -------
    plt.Figure
        The generated plot figure.

    """
    prop_cycle = plt.rcParams["axes.prop_cycle"]
    colors = prop_cycle.by_key()["color"]

    x, y = next(ds.as_numpy_iterator())
    x1, x2 = None, None

    if history_size and len(history_columns) and len(meta_columns):
        x1, x2 = x
    elif history_size and len(history_columns):
        x1 = x
    elif len(meta_columns):
        x2 = x

    x1_columns = sorted(history_columns)
    x2_columns = sorted(meta_columns)
    y_columns = sorted(prediction_columns)

    x1_column_ch = {k: c for c, k in enumerate(x1_columns)}
    y_column_ch = {k: c for c, k in enumerate(y_columns)}

    fig = plt.figure(figsize=figsize)
    if x1 is not None:
        for c in x1_columns:
            ch = x1_column_ch[c]
            plt.plot(
                np.arange(history_size),
                x1[0, ..., ch],
                color=colors[ch % len(colors)],
                label=c,
            )

    if x2 is not None:
        plt.table(
            cellText=x2[0],
            rowLabels=["meta data"],
            colLabels=x2_columns,
            colLoc="right",
            loc="top",
            edges="horizontal",
        )
    for c in y_columns:
        ch = y_column_ch[c]
        color_idx = x1_column_ch.get(c, ch)
        plt.scatter(
            np.arange(prediction_size) + history_size,
            y[0, ..., ch],
            color=colors[color_idx % len(colors)],
            label=c + " (target)",
            s=32,
            edgecolors="k",
        )

    # Adjust layout to make room for the table:
    plt.subplots_adjust(left=0.1, top=1.2)
    plt.legend(loc="lower right")
    plt.xlabel("Time Steps")
    plt.ylabel("Observations")
    plt.tight_layout()

    return fig



# -*- time-stamp-pattern: "changed[\s]+:[\s]+%%$"; -*-
# AUTHOR INFORMATION ##########################################################
# file    : time_series_split.py
# author  : Marcel Arpogaus <marcel dot arpogaus at gmail dot com>
#
# created : 2022-01-07 09:02:38 (Marcel Arpogaus)
# changed : 2024-02-19 12:09:44 (Marcel Arpogaus)
# DESCRIPTION #################################################################
# ...
# LICENSE #####################################################################
# Copyright 2022 Marcel Arpogaus
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###############################################################################


from typing import Any

import numpy as np


class TimeSeriesSplit:
    RIGHT: int = 0
    LEFT: int = 1

    def __init__(self, split_size: float, split: int) -> None:
        """Initialize the TimeSeriesSplit object.

        Parameters
        ----------
        split_size : float
            Proportion of split size for the time series data.
        split : int
            Split position indicator (LEFT or RIGHT).

        """
        self.split_size: float = split_size
        self.split: int = split

    def __call__(self, data: Any) -> Any:
        """Splits the time series data based on the split position.

        Parameters
        ----------
        data : Any
            Input time series data to split.

        Returns
        -------
        Any
            Left or right split of the time series data based on the split position.

        """
        data = data.sort_index()
        days = data.index.date
        days = days.astype("datetime64[m]")
        right = days[int(len(days) * self.split_size)]
        left = right - np.timedelta64(1, "m")
        if self.split == self.LEFT:
            return data.loc[: str(left)]
        else:
            return data.loc[str(right) :]



# -*- time-stamp-pattern: "changed[\s]+:[\s]+%%$"; -*-
# AUTHOR INFORMATION ##########################################################
# file    : patch_generator.py
# author  : Marcel Arpogaus <marcel dot arpogaus at gmail dot com>
#
# created : 2022-01-07 09:02:38 (Marcel Arpogaus)
# changed : 2024-02-19 12:52:06 (Marcel Arpogaus)
# DESCRIPTION #################################################################
# ...
# LICENSE #####################################################################
# Copyright 2022 Marcel Arpogaus
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###############################################################################

import tensorflow as tf


class PatchGenerator:
    """A generator class that creates windows from provided time-series data.

    Attributes
    ----------
    window_size : int
        The size of each patch.
    shift : int
        The shift between patches.

    """

    def __init__(self, window_size: int, shift: int) -> None:
        """Parameters
        ----------
        window_size : int
            The size of each patch.
        shift : int
            The shift between patches.

        """
        self.window_size: int = window_size
        self.shift: int = shift

    def __call__(self, data: tf.Tensor) -> tf.data.Dataset:
        """Converts input data into patches of provided window size.

        Parameters
        ----------
        data : tf.Tensor
            The input data to generate patches from.

        Returns
        -------
        tf.data.Dataset
            A dataset of patches generated from the input data.

        """

        data_set: tf.data.Dataset = tf.data.Dataset.from_tensor_slices(data)

        data_set = data_set.window(
            size=self.window_size,
            shift=self.shift,
            drop_remainder=True,
        ).flat_map(lambda sub: sub.batch(self.window_size, drop_remainder=True))

        return data_set



# -*- time-stamp-pattern: "changed[\s]+:[\s]+%%$"; -*-
# AUTHOR INFORMATION ##########################################################
# file    : __init__.py
# author  : Marcel Arpogaus <marcel dot arpogaus at gmail dot com>
#
# created : 2022-01-07 09:02:38 (Marcel Arpogaus)
# changed : 2024-02-19 13:55:19 (Marcel Arpogaus)
# DESCRIPTION #################################################################
# ...
# LICENSE #####################################################################
# Copyright 2022 Marcel Arpogaus
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###############################################################################
""".. include:: ../../README.md"""



# -*- time-stamp-pattern: "changed[\s]+:[\s]+%%$"; -*-
# AUTHOR INFORMATION ##########################################################
# file    : patch_processor.py
# author  : Marcel Arpogaus <marcel dot arpogaus at gmail dot com>
#
# created : 2022-01-07 09:02:38 (Marcel Arpogaus)
# changed : 2024-02-19 12:41:38 (Marcel Arpogaus)
# DESCRIPTION #################################################################
# ...
# LICENSE #####################################################################
# Copyright 2022 Marcel Arpogaus
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###############################################################################
from typing import Dict, List, Tuple

import tensorflow as tf


class PatchPreprocessor:
    """A class for preprocessing patches of time series data.

    Parameters
    ----------
    history_size : int
        The number of past time steps to include for each sample.
    history_columns : List[str]
        The names of columns in the dataset that are part of the historical data.
    meta_columns : List[str]
        The names of columns in the dataset that are part of the meta data.
    prediction_columns : List[str]
        The names of columns in the dataset that are to be predicted.

    Attributes
    ----------
    history_size : int
        The number of past time steps to include for each sample.
    history_columns : List[str]
        The names of columns in the dataset that are part of the historical data.
    meta_columns : List[str]
        The names of columns in the dataset that are part of the meta data.
    prediction_columns : List[str]
        The names of columns in the dataset that are to be predicted.
    column_idx : Dict[str, int]
        A mapping from column names to their indices in the input dataset.

    """

    def __init__(
        self,
        history_size: int,
        history_columns: List[str],
        meta_columns: List[str],
        prediction_columns: List[str],
    ) -> None:
        self.history_size: int = history_size
        self.history_columns: List[str] = history_columns
        self.meta_columns: List[str] = meta_columns
        self.prediction_columns: List[str] = prediction_columns

        assert (
            len(set(history_columns + meta_columns)) != 0
        ), "No feature columns provided"
        if len(meta_columns) == 0:
            assert history_size > 0, (
                "history_size must be a positive integer greater than zero"
                ", when no meta date is used"
            )
        else:
            assert history_size >= 0, "history_size must be a positive integer"
        assert len(prediction_columns) != 0, "No prediction columns provided"

        columns = sorted(list(set(history_columns + prediction_columns + meta_columns)))
        self.column_idx: Dict[str, int] = {c: i for i, c in enumerate(columns)}

    def __call__(self, patch: tf.Tensor) -> Tuple[tf.Tensor, tf.Tensor]:
        """Process a single patch of data.

        Parameters
        ----------
        patch : tf.Tensor
            A tensor representing a patch of time series data.

        Returns
        -------
        Tuple[tf.Tensor, tf.Tensor]
            A tuple containing the input features (as one or more tensors) and
            the target values as tensors.

        """
        y: List[tf.Tensor] = []
        x_hist: List[tf.Tensor] = []
        x_meta: List[tf.Tensor] = []

        x_columns = sorted(set(self.history_columns + self.meta_columns))
        y_columns = sorted(self.prediction_columns)

        assert (
            len(set(x_columns + y_columns)) == patch.shape[-1]
        ), "Patch shape does not match column number"

        for c in y_columns:
            column = patch[self.history_size :, self.column_idx[c]]
            y.append(column)

        for c in x_columns:
            column = patch[:, self.column_idx[c], None]
            if self.history_size and c in self.history_columns:
                x_hist.append(column[: self.history_size, 0])
            if c in self.meta_columns:
                x_meta.append(column[self.history_size, None, ...])

        y = tf.stack(y, axis=-1)
        x: List[tf.Tensor] = []
        if len(x_hist):
            x.append(tf.stack(x_hist, axis=-1))
        if len(x_meta):
            x.append(tf.concat(x_meta, axis=-1))

        if len(x) > 1:
            return tuple(x), y
        else:
            return x[0], y



from tensorflow_time_series_dataset.loaders import CSVDataLoader


def test_csv_loader(tmp_csv_with_id):
    test_data_path, test_df = tmp_csv_with_id
    csv_loader = CSVDataLoader(test_data_path)
    loaded_df = csv_loader()
    assert loaded_df.equals(test_df.set_index("date_time")), "Error: CSVLoader failed"



# -*- time-stamp-pattern: "changed[\s]+:[\s]+%%$"; -*-
# AUTHOR INFORMATION ##########################################################
# file    : windowed_time_series_dataset_factory.py
# author  : Marcel Arpogaus <marcel dot arpogaus at gmail dot com>
#
# created : 2022-01-07 09:02:38 (Marcel Arpogaus)
# changed : 2024-02-19 12:57:42 (Marcel Arpogaus)
# DESCRIPTION #################################################################
# ...
# LICENSE #####################################################################
# Copyright 2022 Marcel Arpogaus
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###############################################################################
from typing import Any, Dict, List

import tensorflow as tf

from .pipeline import WindowedTimeSeriesPipeline


class WindowedTimeSeriesDatasetFactory:
    """A factory to initialize a new WindowedTimeSeriesPipeline.

    Parameters
    ----------
    history_columns : List[str]
        List of columns representing the history data.
    prediction_columns : List[str]
        List of columns representing the prediction data.
    meta_columns : List[str], optional
        List of columns representing the meta data, by default [].
    dtype : tf.DType, optional
        Data type of the dataset, by default tf.float32.
    **pipeline_kwargs : Any
        Keyword arguments for the data pipeline.

    """

    default_pipeline_kwargs: Dict[str, Any] = {
        "shift": None,
        "batch_size": 32,
        "cycle_length": 1,
        "shuffle_buffer_size": 1000,
        "cache": True,
    }

    def __init__(
        self,
        history_columns: List[str],
        prediction_columns: List[str],
        meta_columns: List[str] = [],
        dtype: tf.DType = tf.float32,
        **pipeline_kwargs: Any,
    ) -> None:
        self.columns: set = set(history_columns + meta_columns + prediction_columns)
        self.preprocessors: List[Any] = []
        self.data_loader: Any = None
        self.dtype: tf.DType = dtype
        self.data_pipeline: WindowedTimeSeriesPipeline = WindowedTimeSeriesPipeline(
            history_columns=history_columns,
            meta_columns=meta_columns,
            prediction_columns=prediction_columns,
            **{**self.default_pipeline_kwargs, **pipeline_kwargs},
        )

    def add_preprocessor(self, preprocessor: Any) -> None:
        """Add a preprocessor to the list of preprocessors.

        Parameters
        ----------
        preprocessor : Any
            The preprocessor function to be added.

        """
        self.preprocessors.append(preprocessor)

    def set_data_loader(self, data_loader: Any) -> None:
        """Set a data loader for the dataset.
        bk.

        Parameters
        ----------
        data_loader : Any
            The data loader function.

        """
        self.data_loader = data_loader

    def get_dataset(self, data: Any = None) -> tf.data.Dataset:
        """Get the dataset for the given data.

        Parameters
        ----------
        data : Any, optional
            The data to create the dataset from, by default None.

        Returns
        -------
        tf.data.Dataset
            The created TensorFlow dataset.

        """
        if self.data_loader is not None and data is None:
            data = self.data_loader()
        elif data is not None:
            data = data
        else:
            raise ValueError("No data provided")

        for preprocessor in self.preprocessors:
            data = preprocessor(data)

        if not isinstance(data, tf.data.Dataset):
            data = data[sorted(self.columns)]
            tensors = tf.convert_to_tensor(data, dtype=self.dtype)
            data = tf.data.Dataset.from_tensors(tensors)

        ds = self.data_pipeline(data)
        return ds

    def __call__(self, data: Any = None) -> tf.data.Dataset:
        """Call method to get the dataset.

        Parameters
        ----------
        data : Any, optional
            The data to create the dataset from, by default None.

        Returns
        -------
        tf.data.Dataset
            The created TensorFlow dataset.

        """
        return self.get_dataset(data)
